#pragma once
#include <vector>
#include <glad/glad.h>
#include <J3ML/LinearAlgebra/Vector2.hpp>
#include <J3ML/LinearAlgebra/Vector2i.hpp>
#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <J3ML/LinearAlgebra/Vector4.hpp>

namespace JGL {
    class VRamList;
}

/// A wrapped for "Vertex Buffer Object" In OpenGL, Store things in VRam.
class JGL::VRamList {
private:
    GLuint list_handle = 0;
    long num_elements = 0;
    bool element_array_buffer = false;
    void load(const GLfloat* data, const long& size);
    void load(const GLuint* data, const long& size);
    void SetData(void* data, const long& length);
    void UpdateData(void* data, const long& offset, const long& length);
    void Erase();
public:
    VRamList(const GLuint* data, const long& length);
    VRamList(const GLfloat* data, const long& length);
    VRamList(const Vector2* data, const long& length);
    VRamList(const Vector3* data, const long& length);
    VRamList(const Vector4* data, const long& length);

    ~VRamList();
    /** Copying around the VBO data to a new VBO like this is slow.
     * Pass to function by const reference or pointer always. */
    VRamList(const VRamList& rhs);
    VRamList() = default;
public:
    [[nodiscard]] GLuint GetHandle() const;
    /// Returns the number of elements in the list.
    [[nodiscard]] long GetLength() const;
    /// Returns the size of the data in bytes.
    [[nodiscard]] size_t GetDataSize() const;
    /** Get VBO data back from the GPU. This is *bad* because the CPU is going to wait
    * for the transfer to finish. Has limited use other than testing. */
    [[nodiscard]] std::vector<GLfloat> GetDataF() const;
    [[nodiscard]] std::vector<GLuint> GetDataUI() const;
    [[nodiscard]] bool IsFloatArray() const;
    /** Replace the data of an existing VBO in it's entirety. Must be same type.
      * "length" refers to the number of elements in data. Not the number of bytes. */
    void SetData(const GLfloat* data, const long& length);
    void SetData(const Vector2* data, const long& length);
    void SetData(const Vector3* data, const long& length);
    void SetData(const Vector4* data, const long& length);

    void SetData(const GLuint* data, const long& length);
    void SetData(const Vector2i* data, const long& length);

    /** Update only a portion of the data in a VBO. Must be same type.
     * "length" refers to the number of elements in data. Not the number of bytes.
     * "offset" refers the number of Typename T into the buffer the data you want to change is.
     * For ex, offset 0 and length of 1 overwrites the first value. Offset 1 the second etc */
     void UpdateData(const GLfloat* data, const long& offset, const long& length);
     void UpdateData(const Vector2* data, const long& offset, const long& length);
     void UpdateData(const Vector3* data, const long& offset, const long& length);
     void UpdateData(const Vector4* data, const long& offset, const long& length);

     void UpdateData(const GLuint* data, const long& offset, const long& length);
     void UpdateData(const Vector2i* data, const long& offset, const long& length);
};