#pragma once

#include <vector>
#include <JGL/types/VRamList.h>
#include <JGL/types/Skeleton.h>
#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <J3ML/Geometry/Sphere.hpp>
#include <J3ML/Geometry/AABB.hpp>
#include <J3ML/Geometry/OBB.hpp>

namespace JGL {
    /// A point that is part of an object in 3D space.
    typedef Vector3 Vertex;
    /// A direction vector which describes which way a triangle is facing.
    typedef Vector3 Normal;
    /// 2D positions that describe how a texture is to be wrapped around a 3D object.
    typedef Vector2 TextureCoordinate;
    /// Container for storing 3D models in v-ram, system memory, or both.
    class VertexArray;
}

class JGL::VertexArray {
private:
    Sphere me_sphere;
    OBB me_obb;
protected:
    virtual void CreateMESphere();
    virtual void CreateMEOBB();
protected:
    std::vector<Animation> animations{};
protected:
    VRamList vertices;
    VRamList indices;
    VRamList normals;
    VRamList texture_coordinates;
protected:
    /** For models which are not animated, This is intended for a low quality version in
     * system memory for calculations to be done on the CPU. For models that are, the default pose of the model is here.
     */
    std::vector<Vertex> local_vertices{};
    std::vector<unsigned int> local_indices{};
    std::vector<TextureCoordinate> local_texture_coordinates{};
    std::vector<Normal> local_normals{};
public:
    /** Don't use these for anything other than drawing because the GPU is gonna spin during read-back */
    [[nodiscard]] VRamList GetVertices() const;
    [[nodiscard]] VRamList GetIndices() const;
    [[nodiscard]] VRamList GetNormals() const;
    [[nodiscard]] VRamList GetTextureCoordinates() const;

    /** These are for cpu side calculations */
    [[nodiscard]] std::vector<Vertex> GetLocalVertices() const;
    [[nodiscard]] std::vector<unsigned int> GetLocalIndices() const;
    [[nodiscard]] std::vector<TextureCoordinate> GetLocalTextureCoordinates() const;
    [[nodiscard]] std::vector<Normal> GetLocalNormals() const;
public:
    /// Returns true if the VertexArray does not have any animations.
    bool Static();
public:
    /// Provides the minimally enclosing bounding sphere of the vertex array given information from the instance.
    /// @param scale The scale of the instance.
    /// @param translate_part The center of the sphere would be shifted in 3D space by this. Primarily for world space.
    [[nodiscard]] Sphere GetMESphere(const Vector3& scale = Vector3::One, const Vector3& translate_part = Vector3::Zero) const;
    /// Provides the minimally enclosing bounding sphere of the vertex array given information from the instance.
    /// @param instance_matrix A Matrix4x4 which contains scale, rotation, and translation.
    /// @param translate Whether or not to translate to world space by the translate part of the Matrix.
    [[nodiscard]] Sphere GetMESphere(const Matrix4x4& instance_matrix, bool translate = false) const;

    /// Provides the minimally enclosing oriented bounding box of the vertex array given information from the instance.
    /// @param rotation_matrix A Matrix3x3 representing rotation in 3D space.
    /// @param scale The scale of the instance.
    /// @param translate_part The center of the box would be shifted in 3D space by this. Primarily for world space.
    [[nodiscard]] OBB GetMEOBB(const Matrix3x3& rotation_matrix, const Vector3& scale = Vector3::One, const Vector3& translate_part = Vector3::Zero) const;
    /// Provides the minimally enclosing oriented bounding box of the vertex array given information from the instance.
    /// @param instance_matrix A Matrix4x4 which contains scale, rotation, and translation.
    /// @param translate Whether or not to translate to world space by the translate part of the Matrix.
    [[nodiscard]] OBB GetMEOBB(const Matrix4x4& instance_matrix, bool translate = false) const;

    /// Provides the minimally enclosing axis-aligned bounding box of the vertex array given information from the instance.
    /// @param rotation_matrix A Matrix3x3 representing rotation in 3D space.
    /// @param scale The scale of the instance.
    /// @param translate_part The center of the box would be shifted in 3D space by this. Primarily for world space.
    [[nodiscard]] AABB GetMEAABB(const Matrix3x3& rotation_matrix, const Vector3& scale = Vector3::One, const Vector3& translate_part = Vector3::Zero) const;
    [[nodiscard]] AABB GetMEAABB(const Matrix4x4& instance_matrix, bool translate = false) const;
public:
    /// Vertices are required, Everything else is optional.
    VertexArray(const Vertex* vertex_positions, const long& vp_length, const unsigned int* vertex_indices = nullptr, const long& vi_length = 0,
                const Normal* vertex_normals = nullptr, const long& vn_length = 0, const TextureCoordinate* texture_coordinates = nullptr, const long& vt_length = 0);

    /// Vertices are required, Everything else is optional.
    explicit VertexArray(const std::vector<Vertex>& vertex_positions, const std::vector<unsigned int>& vertex_indices = {},
                         const std::vector<Normal>& vertex_normals = {}, const std::vector<TextureCoordinate>& texture_coordinates = {});
};

using namespace JGL;
static VertexArray Animate(int animation_id, float animation_time);
static VertexArray Animate(const AnimationState& anim_state);

