#include <JGL/types/VertexArray.h>

using namespace JGL;

VRamList VertexArray::GetVertices() const {
    return vertices;
}

VRamList VertexArray::GetIndices() const {
    return indices;
}

JGL::VRamList VertexArray::GetNormals() const {
    return normals;
}

VRamList VertexArray::GetTextureCoordinates() const {
    return texture_coordinates;
}

std::vector<Vertex> VertexArray::GetLocalVertices() const {
    return local_vertices;
}

std::vector<unsigned int> VertexArray::GetLocalIndices() const {
    return local_indices;
}

std::vector<TextureCoordinate> VertexArray::GetLocalTextureCoordinates() const {
    return local_texture_coordinates;
}

std::vector<Normal> VertexArray::GetLocalNormals() const {
    return local_normals;
}

Sphere VertexArray::GetMESphere(const Vector3& scale, const Vector3& translate_part) const {
    if (scale == Vector3::One)
        return { me_sphere.Position + translate_part, me_sphere.Radius };

    float scale_factor = 0;
    if (scale.x > scale_factor)
        scale_factor = scale.x;
    if (scale.y > scale_factor)
        scale_factor = scale.y;
    if (scale.z > scale_factor)
        scale_factor = scale.z;

    return { me_sphere.Position + translate_part, me_sphere.Radius * scale_factor };
}

Sphere VertexArray::GetMESphere(const Matrix4x4& instance_matrix, bool translate) const {
    Vector3 scale = instance_matrix.GetScale();
    Vector3 origin = Vector3::Zero;
    if (translate)
        origin = instance_matrix.GetTranslatePart();

    return GetMESphere(scale, origin);
}

void VertexArray::CreateMESphere() {
    Sphere result = {Vector3::Zero, 0};
    Vector3 minimum = {std::numeric_limits<float>::max(), std::numeric_limits<float>::max(), std::numeric_limits<float>::max() };
    Vector3 maximum = { std::numeric_limits<float>::lowest(), std::numeric_limits<float>::lowest(), std::numeric_limits<float>::lowest()};

    for (const auto& vertex : local_vertices) {
        if (vertex.x < minimum.x)
            minimum.x = vertex.x;
        if (vertex.x > maximum.x)
            maximum.x = vertex.x;

        if (vertex.y < minimum.y)
            minimum.y = vertex.y;
        if (vertex.y > maximum.y)
            maximum.y = vertex.y;

        if (vertex.z < minimum.z)
            minimum.z = vertex.z;
        if (vertex.z > maximum.z)
            maximum.z = vertex.z;
    }
    result.Position = (minimum + maximum) * 0.5f;
    for (const auto& vertex : local_vertices) {
        float distance_squared = Vector3::DistanceSquared(result.Position, vertex);
        if (distance_squared > result.Radius)
            result.Radius = distance_squared;
    }
    result.Radius = std::sqrt(result.Radius);
    me_sphere = result;
}

void VertexArray::CreateMEOBB() {
    Vector3 minimum(std::numeric_limits<float>::max());
    Vector3 maximum (std::numeric_limits<float>::min());

    for (const auto& vertex : local_vertices) {
        if (vertex.x < minimum.x)
            minimum.x = vertex.x;
        if (vertex.x > maximum.x)
            maximum.x = vertex.x;

        if (vertex.y < minimum.y)
            minimum.y = vertex.y;
        if (vertex.y > maximum.y)
            maximum.y = vertex.y;

        if (vertex.z < minimum.z)
            minimum.z = vertex.z;
        if (vertex.z > maximum.z)
            maximum.z = vertex.z;
    }

    Vector3 position = (minimum + maximum) * 0.5f;
    Vector3 half_extents = (maximum - minimum) * 0.5f;

    me_obb = { position, half_extents, Vector3::UnitX, Vector3::UnitY, Vector3::UnitZ };
}

OBB VertexArray::GetMEOBB(const Matrix3x3& rotation_matrix, const Vector3& scale, const Vector3& translate_part) const {
    Vector3 half_extents = me_obb.r.Mul(scale);
    Vector3 axis0 = rotation_matrix * me_obb.axis[0];
    Vector3 axis1 = rotation_matrix * me_obb.axis[1];
    Vector3 axis2 = rotation_matrix * me_obb.axis[2];
    Vector3 position = me_obb.pos + translate_part;

    return { position, half_extents, axis0, axis1, axis2 };
}

OBB VertexArray::GetMEOBB(const Matrix4x4& instance_matrix, bool translate) const {
    return GetMEOBB(instance_matrix.GetRotatePart(), instance_matrix.GetScale(), translate ? instance_matrix.GetTranslatePart() : Vector3::Zero);
}

AABB VertexArray::GetMEAABB(const Matrix3x3& rotation_matrix, const Vector3& scale, const Vector3& translate_part) const {
    /* kind-of a cheat. It's faster to calculate the oriented bounding box and then put the axis-aligned one around that
     * than to loop over the whole model like an idiot. You gain back all the speed with the simplicity of the physics anyways. - Redacted */
    return GetMEOBB(rotation_matrix, scale, translate_part).MinimalEnclosingAABB();
}

AABB VertexArray::GetMEAABB(const Matrix4x4& instance_matrix, bool translate) const {
    return GetMEAABB(instance_matrix.GetRotatePart(), instance_matrix.GetScale(), translate ? instance_matrix.GetTranslatePart() : Vector3::Zero);
}

VertexArray::VertexArray(const Vector3* vertex_positions, const long& vp_length, const unsigned int* vertex_indices, const long& vi_length,
                         const Normal* vertex_normals, const long& vn_length, const TextureCoordinate* texture_coordinates, const long& vt_length) {

    // TODO decimator. This is a total waste of memory as it sits.
    vertices = VRamList(vertex_positions, vp_length);
    if (vertex_indices && vi_length) {
        indices = VRamList(vertex_indices, vi_length);
        local_indices.resize(vi_length);
        memcpy(local_indices.data(), vertex_indices, sizeof(unsigned int) * vi_length);
    }

    if (vertex_normals && vn_length) {
        normals = VRamList(vertex_normals, vn_length);
        local_normals.resize(vn_length);
        memcpy(local_normals.data(), vertex_indices, sizeof(Normal) * vn_length);
    }

    if (texture_coordinates && vt_length) {
        this->texture_coordinates = VRamList(texture_coordinates, vt_length);
        local_normals.resize(vt_length);
        memcpy(local_texture_coordinates.data(), texture_coordinates, sizeof(TextureCoordinate) * vt_length);
    }

}

VertexArray::VertexArray(const std::vector<Vertex>& vertex_positions, const std::vector<unsigned int>& vertex_indices,
                         const std::vector<Normal>& vertex_normals, const std::vector<TextureCoordinate>& texture_coordinates) {
    vertices = VRamList(vertex_positions.data(), vertex_positions.size());
    local_vertices = vertex_positions;

    if (!vertex_indices.empty()) {
        indices = VRamList(vertex_indices.data(), vertex_indices.size());
        local_indices = vertex_indices;
    }

    if (!vertex_normals.empty()) {
        normals = VRamList(vertex_normals.data(), vertex_normals.size());
        local_normals = vertex_normals;
    }

    if (!texture_coordinates.empty()){
        this->texture_coordinates = VRamList(texture_coordinates.data(), texture_coordinates.size());
        local_texture_coordinates = texture_coordinates;
    }
}

bool VertexArray::Static() {
    return animations.empty();
}
